from enum import Enum


class DiscordConversationMessageMode(Enum):
    """
    The speed with which the message will be shown. A
    fast speed means that it will be shown a short
    period of time.
    """
    
    FAST = 'fast'
    NORMAL = 'normal'
    SLOW = 'slow'


from yta_validation.parameter import ParameterValidator
from dataclasses import dataclass


@dataclass
class DiscordConversationCharacter:

    def __init__(
        self,
        username: str,
        avatar_url: str
    ):
        ParameterValidator.validate_mandatory_string('username', username, do_accept_empty = False)
        ParameterValidator.validate_mandatory_string('avatar_url', avatar_url, do_accept_empty = False)
        
        # TODO: Check if 'avatar_url' is valid

        # TODO: Maybe if they don't provide this information we
        # could auto-generate it
        self.username = username
        self.avatar_url = avatar_url


# TODO: Here we are importing the library itself, but this is not 
# working if we change the library name, no (?)
from yta_image.edition.filter.pixelate import pixelate_image
from yta_general_utils.temp import Temp
from yta_general_utils.file.checker import FileValidator
from yta_validation import PythonValidator
from moviepy import VideoFileClip, ImageSequenceClip
from typing import Union

import shutil


def pixelate_video_file(video_input: Union[VideoFileClip, str], output_filename: str = None):
    """
    This methods pixelates the video progressively and then depixelates
    it progressively again. Finally, it writes the new video as
    'output_filename'.

    TODO: Review this to improve and to check, please, it maybe 
    should be experimental (?)
    """
    # TODO: There is another video pixelating method that I think is
    # faster than this one (the 'artistic.pixelated_video.py')
    if not video_input:
        return None
    
    if PythonValidator.is_string(video_input):
        if not FileValidator.file_is_video_file(video_input):
            return None
        
        video_input = VideoFileClip(video_input)

    PIXELS_AMOUNT_STEP = 64
    
    # We extract frames, and pixelate in ascending order
    original_frames_array = []
    for frame in video_input.iter_frames():
        frame_name = Temp.create_custom_filename('tmp_frame_' + str(len(original_frames_array)) + '.png')
        original_frames_array.append(frame_name)
    video_input.write_images_sequence(Temp.create_custom_filename('tmp_frame_%01d.png'), logger = 'bar')

    # Remove green screen of each frame and store it
    processed_frames_array = []
    num_of_frames = round(video_input.fps * video_input.duration)
    # By now, from start to end
    start_frame = 0
    end_frame = num_of_frames
    pixelate_factor = (1024, 1024)
    pixelation_duration = int(1024 / PIXELS_AMOUNT_STEP) - 2
    # We will start pixelating, maintain, and remove pixelation
    for index, frame in enumerate(original_frames_array):
        tmp_frame_filename = Temp.create_custom_filename('tmp_frame_processed_' + str(index) + '.png')
        processed_frames_array.append(tmp_frame_filename)
        if index >= start_frame and index <= end_frame:
            pixelate_image(frame, pixelate_factor, tmp_frame_filename)

            if index <= (start_frame + pixelation_duration):
                # We are in incrementing process
                tmp_list = list(pixelate_factor)
                tmp_list[0] -= PIXELS_AMOUNT_STEP
                tmp_list[1] -= PIXELS_AMOUNT_STEP
                pixelate_factor = tuple(tmp_list)
            elif index >= (end_frame - pixelation_duration):
                # We are in decrementing process
                tmp_list = list(pixelate_factor)
                tmp_list[0] += PIXELS_AMOUNT_STEP
                tmp_list[1] += PIXELS_AMOUNT_STEP
                pixelate_factor = tuple(tmp_list)
        else:
            # Just copy the original frame
            shutil.copy(original_frames_array[index], tmp_frame_filename)

    clip = ImageSequenceClip(processed_frames_array, fps = video_input.fps).with_audio(video_input.audio)

    if output_filename:
        clip.write_videofile(output_filename)
            
    return clip

    # TODO: Check 'yta_multimedia\video\audio.py' to use ffmpeg as python lib
        
    # TODO: Which one is faster, moviepy or ffmpeg (?)
    #parameters = ['ffmpeg', '-y', '-i', WIP_FOLDER + 'tmp_frame_processed_%01d.png', '-r', '30', '-pix_fmt', 'yuva420p', output_filename]
    #run(parameters)


# TODO: This package is 'position.utils.position' so it doesn't make sense
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.position import Position
from yta_validation import PythonValidator
from typing import Union


"""
    Coords related functions below
"""
def get_moviepy_position(video, background_video, position: Union[Coordinate, Position, tuple]):
    """
    In the process of overlaying and moving the provided 'video' over
    the also provided 'background_video', this method calculates the
    (x, y) tuple position that would be, hypothetically, adapted from
    a 1920x1080 black color background static image. The provided 
    'position' will be transformed into the (x, y) tuple according
    to our own definitions in which the video (that starts in upper left
    corner) needs to be placed to fit the desired 'position'.
    """
    video = VideoParser.to_moviepy(video)
    background_video = VideoParser.to_moviepy(background_video)
    
    # TODO: Maybe simplify these checkings below to a common method (?)
    if not position:
        raise Exception('No "position" provided.')
    
    if not PythonValidator.is_instance(position, [Coordinate, Position, tuple]):
        raise Exception('Provided "position" is not a valid Coordinate nor Position instance nor a tuple.')
    
    if PythonValidator.is_tuple(position) and len(position) != 2:
        # TODO: Maybe apply the normalization limits as limits
        # here for each position tuple element
        raise Exception('Provided "position" is a tuple but does not have 2 values.')
    
    position_tuple = position   # If tuple, here it is
    if PythonValidator.is_instance(position, Position):
        position_tuple = position.get_moviepy_upper_left_corner_tuple(video.size, background_video.size)
    elif PythonValidator.is_instance(position, Coordinate):
        position = position.update_scene_size(background_video.size)
        position_tuple = position.get_moviepy_upper_left_corner_tuple(video.size)

    return position_tuple


from yta_multimedia.video.edition.modifications.video_modification import VideoModification
from yta_multimedia.video.edition.video_editor import VideoEditor
from moviepy.Clip import Clip


class ColorTemperatureVideoModification(VideoModification):
    factor: int = None
    
    def __init__(
        self,
        start_time: float,
        end_time: float,
        layer: int,
        factor: int = 45
    ):
        super().__init__(start_time, end_time, layer)
        
        # TODO: Validate that 'factor' is actually between
        # the limits
        self.factor = factor

    def _modificate(self, video: Clip) -> Clip:
        return VideoEditor(video).change_color_temperature(self.factor)


from yta_multimedia.video.edition.s_effect.s_effect import SEffect, SEffectType
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.objects import WithPositionArgument
from yta_general_utils.math.map.map import Map
from yta_validation import PythonValidator
from typing import Tuple, Union

import numpy as np


class MoveLinearSEffect(SEffect):
    """
    This effect is handled frame by frame and returns
    the array of values for each frame to obtain the
    expected result.
    """
    def __init__(
        self,
        number_of_frames: int,
        initial_position: Union[Coordinate, Position, tuple[int, int]],
        final_position: Union[Coordinate, Position, tuple[int, int]]
    ):
        # TODO: Improve this by first raising Exceptions with 
        # non accepted values and then transforming them
        if PythonValidator.is_instance(initial_position, Coordinate):
            initial_position = initial_position.as_moviepy_tuple
        elif PythonValidator.is_instance(initial_position, Position):
            initial_position = initial_position.get_moviepy_center_tuple()
        elif PythonValidator.is_tuple(initial_position) and len(initial_position) == 2:
            # Default scene of 1920x1080
            initial_position = initial_position
        else:
            raise Exception('The provided "initial_position" parameter is not valid.')

        if PythonValidator.is_instance(final_position, Coordinate):
            final_position = final_position.as_moviepy_tuple
        elif PythonValidator.is_instance(final_position, Position):
            final_position = final_position.get_moviepy_center_tuple() 
        elif PythonValidator.is_tuple(final_position) and len(final_position) == 2:
            # Default scene of 1920x1080
            final_position = final_position
        else:
            raise Exception('The provided "final_position" parameter is not valid.')

        super().__init__(
            number_of_frames,
            SEffectType.MOVEMENT,
            initial_position = initial_position,
            final_position = final_position
        )

    @staticmethod
    def calculate(
        number_of_frames: int,
        initial_position: tuple[int, int],
        final_position: tuple[int, int]
    ) -> Tuple[Union[list[np.ndarray], None], Union[list[int, int], None], Union[list[float, float], None], Union[int, None]]:
        """
        Calculate the values for the 4 arrays:
         
        - frames
        - with_position
        - resized
        - rotated
        
        with the provided 'number_of_frames' and
        additional arguments and return each of
        them if affected, or None if not.
        """
        return (
            None,
            WithPositionArgument(Map().add_coordinate(initial_position).add_coordinate(final_position)).get_values(number_of_frames),
            None,
            None
        )


from yta_multimedia.video.parser import VideoParser
from yta_image.parser import ImageParser
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_validation import PythonValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.dataclasses import FileReturn
from moviepy import ImageClip
from moviepy.Clip import Clip
from typing import Union, Any
import numpy as np


def generate_video_from_image(
    image: Union[str, Any, ImageClip],
    duration: float = 1,
    fps: float = 60,
    output_filename: Union[str, None] = None
) -> Union[FileReturn, Clip]:
    """
    Create an ImageClip of 'duration' seconds with the
    given 'image' and store it locally if 'output_filename'
    is provided.

    This method return a Clip instance if the file is not
    written, and a FileReturn instance if yes.
    """
    ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)

    if not PythonValidator.is_instance(image, ImageClip):
        video = ImageClip(ImageParser.to_numpy(image), duration = duration).with_fps(fps)

    if output_filename:
        video.write_videofile(Output.get_filename(output_filename, FileTypeX.VIDEO))

    return (
        video
        if output_filename is None else
        FileReturn(
            video,
            FileTypeX.VIDEO,
            output_filename
        )
    )

def is_video_transparent(
    video: Clip
):
    """
    Checks if the first frame of the mask of the
    given 'video' has, at least, one transparent
    pixel.
    """
    # We need to detect the transparency from the mask
    video = VideoParser.to_moviepy(video, do_include_mask = True)

    # We need to find, by now, at least one transparent pixel
    # TODO: I would need to check all frames to be sure of this above
    # TODO: The mask can have partial transparency, which 
    # is a value greater than 0, so what do we consider
    # 'transparent' here (?)
    return np.any(video.mask.get_frame(t = 0) == 1)